@@ -1,5 +1,7 @@
 /*
  * Copyright (C) 2010 Igalia S.L.
+ * Copyright (C) 2011 Intel Corporation.
+ *
  *
  * Contact: Iago Toral Quiroga <itoral@igalia.com>
  *
@@ -117,9 +119,9 @@ grl_flickr_plugin_init (GrlPluginRegistry *registry,
                         const GrlPluginInfo *plugin,
                         GList *configs)
 {
-  const gchar *flickr_key;
-  const gchar *flickr_secret;
-  const gchar *flickr_token;
+  gchar *flickr_key;
+  gchar *flickr_secret;
+  gchar *flickr_token;
   GrlConfig *config;
   gboolean public_source_created = FALSE;
 
@@ -157,6 +159,13 @@ grl_flickr_plugin_init (GrlPluginRegistry *registry,
                                            NULL);
     }
 
+    if (flickr_key != NULL)
+      g_free (flickr_key);
+    if (flickr_token != NULL)
+      g_free (flickr_token);
+    if (flickr_secret != NULL)
+      g_free (flickr_secret);
+
     configs = g_list_next (configs);
   }
 
